#include <LocalHistograms.h>
#include <cassert>
#include <cmath>
using namespace std;
using namespace ImageStack;

// We should replace these functions with look-up tables to maximize speed.

float gaussian(float x, float mu, float sigma) {
	float y = (x - mu) / sigma;
	return exp(-.5*y*y) / (sigma * 2.50662827463);
}

float gaussianDerivative(float x, float mu, float sigma) {
	float y = (x - mu) / sigma;
	return (mu - x) * exp(-.5*y*y) / (sigma*sigma*sigma * 2.50662827463);
}

float gaussianIntegral(float x, float mu, float sigma) {
	return .5 * ( 1 + erf((x-mu) / (sigma * 1.41421356237)) );
}

LocalHistograms::LocalHistograms(Image &image, float sigmaIm, float sigmaHist, int nBins, HistogramType type)
	: value(NULL), integral(NULL), derivative(NULL), im(image),
	  imageSigma(sigmaIm), histSigma(sigmaHist), histType(type) {

	assert(im.frames == 1 && im.channels == 1);
	
	Stats imStats(im);
	float minVal = imStats.minimum(0);
	float maxVal = imStats.maximum(0);
	bins.resize(nBins);
	
	float extraSpace = (maxVal-minVal)*0.01 - 1e-5;
	float minBin = minVal - extraSpace;
	float maxBin = maxVal + extraSpace;
	
	for (int i = 0; i < nBins; i++)
		bins[i] = minBin + (maxBin-minBin) * i / (double)(nBins-1);
	
	if (type & HISTOGRAM_VALUE) {
		value = new Image(im.width,im.height,1,nBins);
	
		#pragma omp parallel for
		for (int y = 0; y < im.height; y++)
		for (int x = 0; x < im.width; x++)
		for (int bin = 0; bin < nBins; bin++)
			(*value)(x,y,0)[bin] = gaussian(bins[bin],im(x,y,0)[0],sigmaHist);
		
		FastBlur::apply(*value, sigmaIm, sigmaIm, 0);
	}
	
	if (type & HISTOGRAM_DERIVATIVE) {
		derivative = new Image(im.width,im.height,1,nBins);
	
		#pragma omp parallel for
		for (int y = 0; y < im.height; y++)
		for (int x = 0; x < im.width; x++)
		for (int bin = 0; bin < nBins; bin++)
			(*derivative)(x,y,0)[bin] = gaussianDerivative(bins[bin],im(x,y,0)[0],sigmaHist);
		
		FastBlur::apply(*derivative, sigmaIm, sigmaIm, 0);
	}
	
	if (type & HISTOGRAM_INTEGRAL) {
		integral = new Image(im.width,im.height,1,nBins);
	
		#pragma omp parallel for
		for (int y = 0; y < im.height; y++)
		for (int x = 0; x < im.width; x++)
		for (int bin = 0; bin < nBins; bin++)
			(*integral)(x,y,0)[bin] = gaussianIntegral(bins[bin],im(x,y,0)[0],sigmaHist);
		
		FastBlur::apply(*integral, sigmaIm, sigmaIm, 0);
	}
}

Image LocalHistograms::localMode(int antialiasWidth) {
	Image result(im.width,im.height,1,1);
	
	for (int y = 0; y < im.height; y++)
	for (int x = 0; x < im.width; x++) result(x,y,0)[0] = 0;
	
	assert(derivative);
	
	vector< pair<float,float> > displacements = getDisplacements(antialiasWidth);
	
	#pragma omp parallel for
	for (int y = 0; y < im.height; y++) {
		float *deriv = new float[bins.size()];
		
		for (int x = 0; x < im.width; x++)
		for (unsigned int disp = 0; disp < displacements.size(); disp++) {
			derivative->sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,deriv);
		
			float imValue;
			im.sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,&imValue);
	
			float bestDiff = 1e5;
			float curResult = imValue;
			for (unsigned int bin = 0; bin < bins.size()-1; bin++) {
				float d1 = deriv[bin];
				float d2 = deriv[bin+1];
		
				if (fabs(d1) < 1e-8 && fabs(d2) < 1e-8) continue;
		
				if (d1 > 0 && d2 <= 0) { // peak!
					float frac = -d1 / (d2 - d1);
					float loc = bins[bin] + frac*(bins[bin+1]-bins[bin]);
					float diff = fabs(loc - imValue);
			
					if (diff < bestDiff) {
						curResult = loc;
						bestDiff = diff;
					}
				}
			}
			
			result(x,y,0)[0] += curResult / displacements.size(); // not numerically great...
		}
		delete [] deriv;
	}
	
	return result;
}

Image LocalHistograms::globalMode(int antialiasWidth) {
	Image result(im.width,im.height,1,1);
	
	for (int y = 0; y < im.height; y++)
	for (int x = 0; x < im.width; x++) result(x,y,0)[0] = 0;
	
	assert(derivative);
	assert(value);
	
	vector< pair<float,float> > displacements = getDisplacements(antialiasWidth);
	
	
	#pragma omp parallel for
	for (int y = 0; y < im.height; y++) {
		float *deriv = new float[bins.size()];
		float *val = new float[bins.size()];
		for (int x = 0; x < im.width; x++)
		for (unsigned int disp = 0; disp < displacements.size(); disp++) {
			derivative->sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,deriv);
			value->sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,val);
		
			float imValue;
			im.sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,&imValue);
	
			float bestHeight = -1;
			float curResult = imValue;
			for (unsigned int bin = 0; bin < bins.size()-1; bin++) {
				float d1 = deriv[bin];
				float d2 = deriv[bin+1];
		
				if (fabs(d1) < 1e-8 && fabs(d2) < 1e-8) continue;
		
				if (d1 > 0 && d2 <= 0) { // peak!
					float frac = -d1 / (d2 - d1);
					float loc = bins[bin] + frac*(bins[bin+1]-bins[bin]);
					float height = val[bin] + frac*(val[bin+1]-val[bin]);
			
					if (bestHeight < height) {
						curResult = loc;
						bestHeight = height;
					}
				}
			}
		
			result(x,y,0)[0] += curResult / displacements.size(); // not numerically great...
		}
		delete [] deriv;
		delete [] val;
	}
	
	return result;
}

Image LocalHistograms::median(float target, int antialiasWidth) {
	Image result(im.width,im.height,1,1);
	
	for (int y = 0; y < im.height; y++)
	for (int x = 0; x < im.width; x++) result(x,y,0)[0] = 0;
	
	assert(integral);
	
	vector< pair<float,float> > displacements = getDisplacements(antialiasWidth);
	
	#pragma omp parallel for
	for (int y = 0; y < im.height; y++) {
		float *integralVals = new float[bins.size()];
		for (int x = 0; x < im.width; x++)
		for (unsigned int disp = 0; disp < displacements.size(); disp++) {
			integral->sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,integralVals);
		
			float imValue;
			im.sample2DLinear(x+displacements[disp].first,y+displacements[disp].second,&imValue);
	
			float curResult = imValue;
			for (unsigned int bin = 0; bin < bins.size()-1; bin++) {
				float v1 = integralVals[bin];
				float v2 = integralVals[bin+1];
		
				if (v1 < target && v2 >= target) { // peak!
					float frac = (target-v1)/(v2-v1);
					curResult = bins[bin] + frac*(bins[bin+1]-bins[bin]);
					break;
				}
			}
		
			result(x,y,0)[0] += curResult / displacements.size(); // not numerically great...
		}
		delete [] integralVals;
	}
	
	return result;
}

vector< pair<float,float> > LocalHistograms::getDisplacements(int antialiasWidth) {
	vector< pair<float,float> > result(antialiasWidth*antialiasWidth);
	
	// simple box filter centered within the pixel
	int pos = 0;
	for (int i = 1; i <= antialiasWidth; i++)
	for (int j = 1; j <= antialiasWidth; j++) {
		result[pos].first = (float)i / (float)(antialiasWidth+1) - .5;
		result[pos++].second = (float)j / (float)(antialiasWidth+1) - .5;
	}
	
	return result;
}
